/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2009, 2010 PRO ACTIVE SAS - www.proactive-group.com
   
  Ref_15693.c
  -----------

  This is the reference applications that shows how to work with
  ISO/IEC 15693 cards.
  
  LTX 25/08/2009 : initial release
	JDA 15/12/2010 : fixed everything...

*/
#include "products/springprox/springprox.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void DecodeUid(const BYTE uid[8]);
void GetSystemInformation(void);

WORD block_size  = 0;
WORD block_count = 0;
BOOL has_read_multiple = TRUE;
BOOL is_writable = TRUE;

int main(int argc, char **argv)
{
  SWORD rc;
  int i = 0;
  BOOL fast = FALSE;
	BOOL first = TRUE;
  TCHAR s_buffer[64];
	BYTE buffer[256];
  DWORD features;
  BYTE uid[8];
	WORD block, blocks, length;
	BOOL success = TRUE;
  
  char *device = NULL;
  if (argc > 1)
    device = argv[1];

  printf("SpringCard SDK for K531/SpringProx-CF/CSB Legacy\n");
  printf("\n");
  printf("ISO 15693 reference demo\n");
  printf("-------------------------------------\n");
  printf("www.springcard.com\n\n");

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

  /* Open reader */
  /* ----------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  printf("Reader found\n");
  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  rc = SPROX_ReaderGetFeatures(&features);
  if ((rc == MI_OK) && !(features & 0x00000002))
  {
    printf("Warning: this reader is not 15693-enabled\n");
    printf("         all subsequent tests will fail !\n");
  }

  printf("\n------------------------------\n");

  /* Configure reader in ISO 15693 standard mode */
  /* ------------------------------------------- */
  rc = SPROX_SetConfig(CFG_MODE_ISO_15693_STD);
  if (rc != MI_OK)
  {
    printf("Failed to configure the reader for ISO 15693 operation\n");
    goto close;
  }

  /* RF field ON */
  SPROX_ControlRF(TRUE);

  /* Select any ISO 15693 tag in the RF field */
  /* ---------------------------------------- */
  
  printf("Put your ISO 15693 tag on the reader...\n");

  for (;;)
  {
	  rc = SPROX_Iso15693_SelectAny(fast, uid);
	  if (rc == MI_OK)
	  {
      printf("Found ISO 15693 tag in RF field\n");
      break;
	  }

	  if (i == 5)
	  {
	    fast = TRUE;
	  } else if (i == 10)
	  {
	    fast = FALSE;
	    i=0;
	  }
	  i++;
  }

again:

  printf("Tag found: UID=");
  for (i = 0; i < 8; i++)
    printf("%02X", uid[i]);
  printf(" ; mode=%s\n", fast ? "fast" : "standard");

	DecodeUid(uid);

	GetSystemInformation();

  if (block_count || block_size)
	{
    printf("Reading %d blocks of %dB...\n", block_count, block_size);
	} else
	{
    printf("Trying to read up to 256 blocks\n");
		block_count = 256;
		block_size  = 256;
	}

	for (block = 0; block < block_count; block++)
	{
	  printf("%03d : ", block);
	  length = sizeof(buffer);
    rc = SPROX_Iso15693_ReadSingleBlock(NULL, (BYTE) block, buffer, &length);
    if (rc == MI_OK)
		{
		  for (i=0; i<length; i++)
			{
        printf("%02X", buffer[i]);
			}
			printf("   ");
		  for (i=0; i<length; i++)
			{
        printf("%c", (buffer[i] >= ' ') ? buffer[i] : '.');
			}
			printf("\n");

			if (block_size != length)
			{
			  block_size = length;
			  printf("Actual block size = %d\n", block_size);
			}
		} else
		{
      printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
			if ((rc == MI_ACCESSERR) || (rc == MI_OTHERERR))
			{
			  /* End of tag reached! */
				block_count = block;
				printf("Actual number of blocks = %d\n", block_count);
			  break;
			}
			success = FALSE;
		}
	}

  if (has_read_multiple)
  {
	  printf("Trying the ReadMultiple function\n");

    for (blocks = 1; blocks < block_count; blocks++)
		{
		  if ((blocks * block_size) > 256) break;

		  printf("%03d block(s) (%03dB)... ", blocks, blocks * block_size);
      length = sizeof(buffer);
      rc = SPROX_Iso15693_ReadMultipleBlocks(NULL, 0, (BYTE) blocks, buffer, &length);
			if (rc == MI_OK)
			{
			  if (length == blocks * block_size)
			    printf("OK\n");
				else
				  printf("Got %03dB ?\n", length);
			} else
			{
        printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
				success = FALSE;
			}
		}
  }

	if (is_writable)
	{
	  blocks = block_count;
	  if (blocks > 6) blocks = 6;

    printf("Trying to write %d blocks\n", blocks);

		memset(buffer, 0, sizeof(buffer));
		for (block = 0; block < blocks; block++)
		{
      rc = SPROX_Iso15693_WriteBlock(NULL, (BYTE) block, buffer, block_size);
			if (rc != MI_OK)
			  printf("Pre-erasing block %d: %s (%d)\n", block, SPROX_GetErrorMessage(rc), rc);
		}

		for (i=0; i<sizeof(buffer); i++)
		  buffer[i] = block_size ^ block_count ^ i;

		for (block = 0; block < blocks; block++)
		{
      rc = SPROX_Iso15693_WriteBlock(NULL, (BYTE) block, &buffer[block * block_size], block_size);
			if (rc != MI_OK)
			  printf("Writing block %d: %s (%d)\n", block, SPROX_GetErrorMessage(rc), rc);
		}
		memset(buffer, 0xCC, sizeof(buffer));

		if (has_read_multiple)
		{
		  length = sizeof(buffer);
      rc = SPROX_Iso15693_ReadMultipleBlocks(NULL, 0, (BYTE) blocks, buffer, &length);
			if (rc != MI_OK)
			  printf("Reading %d blocks: %s (%d)\n", blocks, SPROX_GetErrorMessage(rc), rc);
		} else
		{
		  for (block = 0; block < blocks; block++)
			{
			  length = sizeof(buffer);
        rc = SPROX_Iso15693_ReadSingleBlock(NULL, (BYTE) block, &buffer[block * block_size], &length);
			}
			if (rc != MI_OK)
			  printf("Reading block %d: %s (%d)\n", block, SPROX_GetErrorMessage(rc), rc);
		}

		for (i=0; i<(blocks * block_size); i++)
		{
		  if (buffer[i] != (block_size ^ block_count ^ i))
			{
        printf("Write failed at offset %d\n", i);
				break;
			}
		}
		if (i<(blocks * block_size))
		{
			for (i=0; i<(blocks * block_size); i++)
			{
				printf("%02X", block_size ^ block_count ^ i);
			}
			printf("\n");
			for (i=0; i<(blocks * block_size); i++)
			{
        printf("%02X", buffer[i]);
      }
			printf("\n");
			success = FALSE;
		} else
		{
      printf("Write is OK!\n");
		}

		memset(buffer, 0, sizeof(buffer));
		for (block = 0; block < blocks; block++)
		{
      rc = SPROX_Iso15693_WriteBlock(NULL, (BYTE) block, buffer, block_size);
			if (rc != MI_OK)
			  printf("Post-erasing block %d: %s (%d)\n", block, SPROX_GetErrorMessage(rc), rc);
		}
		memset(buffer, 0xCC, sizeof(buffer));

		if (has_read_multiple)
		{
		  length = sizeof(buffer);
      rc = SPROX_Iso15693_ReadMultipleBlocks(NULL, 0, (BYTE) blocks, buffer, &length);
			if (rc != MI_OK)
			{
			  length = sizeof(buffer);
			  printf("Reading %d blocks: %s (%d)\n", blocks, SPROX_GetErrorMessage(rc), rc);
			}
		} else
		{
		  for (block = 0; block < blocks; block++)
        rc = SPROX_Iso15693_ReadSingleBlock(NULL, (BYTE) block, &buffer[block * block_size], &length);
			if (rc != MI_OK)
			  printf("Reading block %d: %s (%d)\n", block, SPROX_GetErrorMessage(rc), rc);
		}

		for (i=0; i<(blocks * block_size); i++)
		{
		  if (buffer[i])
			{
        printf("Post-erasing failed at offset %d\n", i);
				break;
			}
		}
		if (i<(blocks * block_size))
		{
			for (i=0; i<(blocks * block_size); i++)
			{
        printf("%02X", buffer[i]);
      }
			printf("\n");
			success = FALSE;
		} else
		{
      printf("Erase is OK!\n");
		}
	}
 
  SPROX_ControlRF(FALSE);

	if (first && success)
	{
	  printf("Reconnecting...\n");
	  first = FALSE;
    SPROX_ControlRF(TRUE);

		fast = !fast;

	  rc = SPROX_Iso15693_SelectAny(fast, uid);
	  if (rc == MI_OK)
	  {
      goto again;
	  }

    printf("Tag not available in %s fast\n", fast ? "fast" : "standard");
		if (rc != MI_OK)
		{
      printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
		}
	}

	if (success)
	{
    printf("All test successfully passed!\n");
	} else
	{
    printf("Some tests have failed!\n");
	}

close:
  /* Close the reader */
  SPROX_ReaderClose();

done:
  printf("Done\n");
  return 0;
}

void DecodeUid(const BYTE uid[8])
{
  switch (uid[1])
  {
		case 0x02 :
		  printf("Manufacturer:   ST MicroElectronics\n");
			switch ((uid[2] & 0xFC) >> 2)
			{
				case 0x05 :	/* LRI64 : 15 blocks, 1 byte in each block (block 0 et 7 -> UID, block 8 -> AFI, block 9 -> DSFID)*/
										printf("Chip:           LRI64\n");
										has_read_multiple = FALSE;
										is_writable = FALSE;
										break;
				default   :	/* Other ST card */
										printf("Unknown chip!\n");
										break;
			}
		break;

    case 0x04 :
		  printf("Manufacturer:   NXP\n");
			switch (uid[2])
			{
				case 0x01 : /* ICODE SLI : 28 blocks, 4 bytes in each block */
										printf("Chip:           ICODE-SLI\n");
										break;
				case 0x03 : /* ICODE SLI-L : 8 blocks, 4 bytes in each block */
										printf("Chip:           ICODE-SLI-L\n");
										has_read_multiple = FALSE;
										break;
				case 0x02 : /* ICODE SLI-S : 40 blocks, 4 bytes in each block */
										printf("Chip:           ICODE-SLI-S\n");
										has_read_multiple = FALSE;
										break;
				default   : /* Other NXP card */
										printf("Unknown chip!\n");
										break;
			}
			break;

		case 0x07 :
		  printf("Manufacturer:   Texas Instrument\n");
			switch (uid[2] & 0xFE)
			{
				case 0x00 :	/* Tag-it HF-I Plus Inlay : 64 blocks, 4 bytes in each block */
										printf("Chip:           TagIT HF-I Plus (inlay)\n");
										break;
				case 0x80 :	/* Tag-it HF-I Plus Chip : 64 blocks, 4 bytes in each block */
										printf("Chip:           TagIT HF-I Plus (die)\n");
										break;
				case 0xC0 :	/* Tag-it HF-I Standard Chip/Inlay : 11 blocks, 4 bytes in each block (block 8 & 9 -> UID, block 10 -> AFI)*/
                    printf("Chip:           TagIT HF-I Standard\n");
										has_read_multiple = FALSE;
										break;
				case 0xC4 :	/* Tag-it HF-I Pro Chip/Inlay : 12 blocks, 4 bytes in each block (block 8 & 9 -> UID, block 10 -> AFI, block 11 -> password)*/
                    printf("Chip:           TagIT HF-I Pro\n");
										has_read_multiple = FALSE;
										break;
				default   :	/* Other Tag-it card */
										printf("Unknown chip!\n");
										break;
			}
			break;


		case 0x16 :   /* EMMarin */
		  printf("Manufacturer:   ElectronicMarin (or Legic)\n");
			break;

		default   :
		  printf("Unknown manufacturer/unknown chip\n");
			break;
  }
}

void GetSystemInformation(void)
{
  SWORD rc;
  int i = 0;
  BYTE data[256];
  WORD len = sizeof(data);
  
  printf("GetSystemInformation ");
  rc = SPROX_Iso15693_GetSystemInformation(NULL, data, &len);
  if (rc == MI_OK) 
  {
	  printf("[OK]\n");

    if (len > 9)
    {
      printf("UID=");
	    for (i = 1; i < 9; i++)
		    printf("%02X", data[9-i]);
      printf("\n");

      if (data[0] & 0x01)
      {
        printf("DSFID=%02X\n", data[i++]);
      } 
                  
      if (data[0] & 0x02)
      {
        printf("AFI=%02X\n", data[i++]);
      } 
                  
      if (data[0] & 0x04)
      {
			  block_count = data[i++] + 1;
        printf("Number of Blocks=%d\n", block_count);
				block_size = (data[i++] & 0x1F) + 1;
        printf("Block Size=%d\n", block_size);
      }
                  
      if (data[0] & 0x08)
      {
        printf("IC-Ref=%02X\n", data[i++]);
      }
    }
  } else
  {
	  printf("[Failed]\n");
    printf("  %s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }
}

