/*
 * SpringProx or K531 SDK :: direct/hal_win32.c
 * --------------------------------------------
 *
 * Windows specific part
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include "ref_direct.h"

#ifdef WIN32

/*
 * Code for accessing serial port under Windows
 * Adapt this tor your target
 */

#include <windows.h>

HANDLE CommHandle = INVALID_HANDLE_VALUE;

int SerialOpen(char *DeviceName)
{
  DCB dcb;
  COMMTIMEOUTS tmo;

  SerialClose();

  /* Open the communication device */
  CommHandle = CreateFile(DeviceName,
                          GENERIC_READ | GENERIC_WRITE,
                          0, NULL, OPEN_EXISTING, 0, NULL);

  if (CommHandle == INVALID_HANDLE_VALUE)
    return E_COMM_ERROR; /* Communication error */

  /* Setup communication : 38400bps, 8 data, 1 stop, no parity, no flow control */
  GetCommState(CommHandle, &dcb);
  dcb.BaudRate = CBR_38400;
  dcb.fBinary = TRUE;
  dcb.fParity = FALSE;
  dcb.fOutxCtsFlow = FALSE;
  dcb.fOutxDsrFlow = FALSE;
  dcb.fDsrSensitivity = FALSE;
  dcb.fOutX = FALSE;
  dcb.fInX = FALSE;
  dcb.fNull = FALSE;
  dcb.ByteSize = 8;
  dcb.Parity = NOPARITY;
  dcb.StopBits = ONESTOPBIT;
  dcb.fRtsControl = RTS_CONTROL_DISABLE;
  dcb.fDtrControl = DTR_CONTROL_DISABLE;
  dcb.fAbortOnError = FALSE;
  dcb.fTXContinueOnXoff = TRUE;
  SetCommState(CommHandle, &dcb);

  /* Define timeouts with optimal value for easy K531 communication */
  tmo.ReadIntervalTimeout         = 50;
  tmo.ReadTotalTimeoutConstant    = 500;
  tmo.ReadTotalTimeoutMultiplier  = 2;
  tmo.WriteTotalTimeoutConstant   = 100;
  tmo.WriteTotalTimeoutMultiplier = 10;
  SetCommTimeouts(CommHandle, &tmo);

  /* Success */
  return 0;
}

void SerialClose(void)
{
  if (CommHandle != INVALID_HANDLE_VALUE)
    CloseHandle(CommHandle);
  CommHandle = INVALID_HANDLE_VALUE;
}

int SerialSend(unsigned char *buffer, int length)
{
  DWORD dwWritten = 0;

  if (!WriteFile(CommHandle, buffer, length, &dwWritten, 0))
    return E_COMM_ERROR; /* Communication error */

  return (int) dwWritten;
}

int SerialRecv(unsigned char *buffer, int length)
{
  DWORD dwRead = 0;

  if (!ReadFile(CommHandle, buffer, length, &dwRead, 0))
    return E_COMM_ERROR; /* Communication error */

  return (int) dwRead;
}

#endif /* WIN32 */
